# The accounting application of the Mia project.
#   by imacat <imacat@mail.imacat.idv.tw>, 2020/6/29

#  Copyright (c) 2020 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The data models of the accounting application.

"""
import datetime
import re
from decimal import Decimal
from typing import Dict, List, Optional, Mapping

from dirtyfields import DirtyFieldsMixin
from django.db import models
from django.db.models import Q, Max
from django.http import HttpRequest

from mia_core.models import L10nModel, LocalizedModel, StampedModel, \
    RandomPkModel


class Account(DirtyFieldsMixin, LocalizedModel, StampedModel, RandomPkModel):
    """An account."""
    parent = models.ForeignKey(
        "self", on_delete=models.PROTECT, null=True,
        related_name="child_set")
    code = models.CharField(max_length=5, unique=True)
    title_l10n = models.CharField(max_length=32, db_column="title")
    CASH = "1111"
    ACCUMULATED_BALANCE = "3351"
    NET_CHANGE = "3353"

    def __init__(self, *args, **kwargs):
        if "title" in kwargs:
            self.title = kwargs["title"]
            del kwargs["title"]
        super().__init__(*args, **kwargs)
        self.url = None
        self.debit_amount = None
        self.credit_amount = None
        self.amount = None
        self.is_for_debit = None
        self.is_for_credit = None
        self._is_in_use = None
        self._is_parent_and_in_use = None

    def __str__(self):
        """Returns the string representation of this account."""
        return self.code.__str__() + " " + self.title

    def save(self, force_insert=False, force_update=False, using=None,
             update_fields=None):
        self.parent = None if len(self.code) == 1\
            else Account.objects.get(code=self.code[:-1])
        super().save(force_insert=force_insert, force_update=force_update,
                     using=using, update_fields=update_fields)

    @property
    def title(self) -> str:
        return self.get_l10n("title")

    @title.setter
    def title(self, value):
        self.set_l10n("title", value)

    @property
    def option_data(self) -> Dict[str, str]:
        """The data as an option."""
        return {
            "code": self.code,
            "title": self.title,
        }

    @property
    def is_parent_and_in_use(self) -> bool:
        """Whether this is a parent account and is in use."""
        if self._is_parent_and_in_use is None:
            self._is_parent_and_in_use = self.child_set.count() > 0\
                                         and self.record_set.count() > 0
        return self._is_parent_and_in_use

    @is_parent_and_in_use.setter
    def is_parent_and_in_use(self, value: bool) -> None:
        self._is_parent_and_in_use = value

    @property
    def is_in_use(self) -> bool:
        """Whether this account is in use."""
        if self._is_in_use is None:
            self._is_in_use = self.child_set.count() > 0\
                              or self.record_set.count() > 0
        return self._is_in_use

    @is_in_use.setter
    def is_in_use(self, value: bool) -> None:
        self._is_in_use = value


class AccountL10n(DirtyFieldsMixin, L10nModel, StampedModel, RandomPkModel):
    """The localization content of an account."""
    master = models.ForeignKey(
        Account, on_delete=models.CASCADE, related_name="l10n_set")


class Transaction(DirtyFieldsMixin, StampedModel, RandomPkModel):
    """An accounting transaction."""
    date = models.DateField()
    ord = models.PositiveSmallIntegerField(default=1)
    notes = models.CharField(max_length=128, null=True, blank=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._records = None
        self._is_balanced = None
        self._has_order_hole = None
        self.old_date = None

    def __str__(self):
        """Returns the string representation of this accounting
        transaction."""
        return self.date.__str__() + " #" + self.ord.__str__()

    def is_dirty(self, check_relationship=False, check_m2m=None) -> bool:
        """Returns whether the data of this transaction is changed and need
        to be saved into the database.

        Returns:
            True if the data of this transaction is changed and need to be
            saved into the database, or False otherwise.
        """
        if super().is_dirty(check_relationship=check_relationship,
                            check_m2m=check_m2m):
            return True
        if len([x for x in self.records
                if x.is_dirty(check_relationship=check_relationship,
                              check_m2m=check_m2m)]) > 0:
            return True
        kept = [x.pk for x in self.records]
        if len([x for x in self.record_set.all() if x.pk not in kept]) > 0:
            return True
        return False

    def save(self, force_insert=False, force_update=False, using=None,
             update_fields=None):
        # When the date is changed, the orders of the transactions in the same
        # day need to be reordered
        txn_to_sort = []
        if self.date != self.old_date:
            if self.old_date is not None:
                txn_same_day = list(
                    Transaction.objects
                    .filter(Q(date=self.old_date), ~Q(pk=self.pk))
                    .order_by("ord"))
                for i in range(len(txn_same_day)):
                    if txn_same_day[i].ord != i + 1:
                        txn_to_sort.append([txn_same_day[i], i + 1])
            max_ord = Transaction.objects\
                .filter(date=self.date)\
                .aggregate(max=Max("ord"))["max"]
            self.ord = 1 if max_ord is None else max_ord + 1
        # Collects the records to be deleted
        to_keep = [x.pk for x in self.records if x.pk is not None]
        to_delete = [] if self.pk is None \
            else [x for x in self.record_set.all() if x.pk not in to_keep]
        to_save = [x for x in self.records
                   if x.is_dirty(check_relationship=True)]
        for record in to_save:
            record.current_user = self.current_user
        # Runs the update
        super().save(force_insert=force_insert, force_update=force_update,
                     using=using, update_fields=update_fields)
        for record in to_delete:
            record.delete()
        for record in to_save:
            record.save(force_insert=force_insert,
                        force_update=force_update,
                        using=using, update_fields=update_fields)
        for x in txn_to_sort:
            Transaction.objects.filter(pk=x[0].pk).update(ord=x[1])

    def delete(self, using=None, keep_parents=False):
        txn_same_day = list(
            Transaction.objects
            .filter(Q(date=self.date), ~Q(pk=self.pk))
            .order_by("ord"))
        txn_to_sort = []
        for i in range(len(txn_same_day)):
            if txn_same_day[i].ord != i + 1:
                txn_to_sort.append([txn_same_day[i], i + 1])
        Record.objects.filter(transaction=self).delete()
        super().delete(using=using, keep_parents=keep_parents)
        for x in txn_to_sort:
            Transaction.objects.filter(pk=x[0].pk).update(ord=x[1])

    def fill_from_post(self, post: Dict[str, str], request: HttpRequest,
                       txn_type: str):
        """Fills the transaction from the POST data.  The POST data must be
        validated and clean at this moment.

        Args:
            post: The POST data.
            request: The request.
            txn_type: The transaction type.
        """
        self.old_date = self.date
        m = re.match(r"^(\d{4})-(\d{2})-(\d{2})$", post["date"])
        self.date = datetime.date(
            int(m.group(1)),
            int(m.group(2)),
            int(m.group(3)))
        self.notes = post.get("notes")
        # The records
        max_no = self._find_max_record_no(txn_type, post)
        records = []
        for record_type in max_no.keys():
            for i in range(max_no[record_type]):
                no = i + 1
                if F"{record_type}-{no}-id" in post:
                    record = Record.objects.get(
                        pk=post[F"{record_type}-{no}-id"])
                else:
                    record = Record(
                        is_credit=(record_type == "credit"),
                        transaction=self)
                record.ord = no
                record.account = Account.objects.get(
                    code=post[F"{record_type}-{no}-account"])
                if F"{record_type}-{no}-summary" in post:
                    record.summary = post[F"{record_type}-{no}-summary"]
                else:
                    record.summary = None
                record.amount = Decimal(post[F"{record_type}-{no}-amount"])
                records.append(record)
        if txn_type != "transfer":
            if txn_type == "expense":
                if len(self.credit_records) > 0:
                    record = self.credit_records[0]
                else:
                    record = Record(is_credit=True, transaction=self)
            else:
                if len(self.debit_records) > 0:
                    record = self.debit_records[0]
                else:
                    record = Record(is_credit=False, transaction=self)
            record.ord = 1
            record.account = Account.objects.get(code=Account.CASH)
            record.summary = None
            record.amount = sum([x.amount for x in records])
            records.append(record)
        self.records = records
        self.current_user = request.user

    @staticmethod
    def _find_max_record_no(txn_type: str,
                            post: Mapping[str, str]) -> Dict[str, int]:
        """Finds the max debit and record numbers from the POSTed form.

        Args:
            txn_type: The transaction type.
            post: The POSTed data.

        Returns:
            The max debit and record numbers from the POSTed form.
        """
        max_no = {}
        if txn_type != "credit":
            max_no["debit"] = 0
        if txn_type != "debit":
            max_no["credit"] = 0
        for key in post.keys():
            m = re.match(
                (r"^(debit|credit)-([1-9]\d*)-"
                 r"(id|ord|account|summary|amount)$"),
                key)
            if m is None:
                continue
            record_type = m.group(1)
            if record_type not in max_no:
                continue
            no = int(m.group(2))
            if max_no[record_type] < no:
                max_no[record_type] = no
        return max_no

    @property
    def records(self):
        """The records of the transaction.

        Returns:
            List[Record]: The records.
        """
        if self._records is None:
            if self.pk is None:
                self._records = []
            else:
                self._records = list(self.record_set.all())
                self._records.sort(key=lambda x: (x.is_credit, x.ord))
        return self._records

    @records.setter
    def records(self, value):
        self._records = value

    @property
    def debit_records(self):
        """The debit records of this transaction.

        Returns:
            List[Record]: The records.
        """
        return [x for x in self.records if not x.is_credit]

    def debit_total(self) -> Decimal:
        """The total amount of the debit records."""
        return sum([x.amount for x in self.debit_records
                    if isinstance(x.amount, Decimal)])

    @property
    def debit_summaries(self) -> List[str]:
        """The summaries of the debit records."""
        return [x.account.title if x.summary is None else x.summary
                for x in self.debit_records]

    @property
    def credit_records(self):
        """The credit records of this transaction.

        Returns:
            List[Record]: The records.
        """
        return [x for x in self.records if x.is_credit]

    def credit_total(self) -> Decimal:
        """The total amount of the credit records."""
        return sum([x.amount for x in self.credit_records
                    if isinstance(x.amount, Decimal)])

    @property
    def credit_summaries(self) -> List[str]:
        """The summaries of the credit records."""
        return [x.account.title if x.summary is None else x.summary
                for x in self.credit_records]

    @property
    def amount(self) -> Decimal:
        """The amount of this transaction."""
        return self.debit_total()

    @property
    def is_balanced(self) -> bool:
        """Whether the sum of the amounts of the debit records is the
        same as the sum of the amounts of the credit records. """
        if self._is_balanced is None:
            debit_sum = sum([x.amount for x in self.debit_records])
            credit_sum = sum([x.amount for x in self.credit_records])
            self._is_balanced = debit_sum == credit_sum
        return self._is_balanced

    @is_balanced.setter
    def is_balanced(self, value: bool) -> None:
        self._is_balanced = value

    def has_many_same_day(self) -> bool:
        """whether there are more than one transactions at this day,
        so that the user can sort their orders. """
        return Transaction.objects.filter(date=self.date).count() > 1

    @property
    def has_order_hole(self) -> bool:
        """Whether the order of the transactions on this day is not
        1, 2, 3, 4, 5..., and should be reordered. """
        if self._has_order_hole is None:
            orders = [x.ord for x in Transaction.objects.filter(
                date=self.date)]
            if len(orders) == 0:
                self._has_order_hole = False
            if max(orders) != len(orders):
                self._has_order_hole = True
            elif min(orders) != 1:
                self._has_order_hole = True
            elif len(orders) != len(set(orders)):
                self._has_order_hole = True
            else:
                self._has_order_hole = False
        return self._has_order_hole

    @has_order_hole.setter
    def has_order_hole(self, value: bool) -> None:
        self._has_order_hole = value

    @property
    def is_cash_income(self) -> bool:
        """Whether this transaction is a cash income transaction."""
        debit_records = self.debit_records
        return (len(debit_records) == 1
                and debit_records[0].account.code == Account.CASH
                and debit_records[0].summary is None)

    @property
    def is_cash_expense(self) -> bool:
        """Whether this transaction is a cash expense transaction."""
        credit_records = self.credit_records
        return (len(credit_records) == 1
                and credit_records[0].account.code == Account.CASH
                and credit_records[0].summary is None)

    @property
    def type(self) -> str:
        """The transaction type."""
        if self.is_cash_expense:
            return "expense"
        elif self.is_cash_income:
            return "income"
        else:
            return "transfer"


class Record(DirtyFieldsMixin, StampedModel, RandomPkModel):
    """An accounting record."""
    transaction = models.ForeignKey(
        Transaction, on_delete=models.CASCADE)
    is_credit = models.BooleanField()
    ord = models.PositiveSmallIntegerField()
    account = models.ForeignKey(
        Account, on_delete=models.PROTECT)
    summary = models.CharField(max_length=128, blank=True, null=True)
    amount = models.DecimalField(max_digits=18, decimal_places=2)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._debit_amount: Optional[Decimal] = None
        self._credit_amount: Optional[Decimal] = None
        self.balance: Optional[Decimal] = None
        self._is_balanced = None
        self._has_order_hole = None
        self._is_payable = None
        self._is_existing_equipment = None
        self.is_payable = False
        self.is_existing_equipment = False

    def __str__(self):
        """Returns the string representation of this accounting
        record."""
        return "%s %s %s %s" % (
            self.transaction.date,
            self.account.title,
            self.summary,
            self.amount)

    @property
    def debit_amount(self) -> Optional[Decimal]:
        """The debit amount of this accounting record."""
        if self._debit_amount is None:
            self._debit_amount = self.amount if not self.is_credit else None
        return self._debit_amount

    @debit_amount.setter
    def debit_amount(self, value: Optional[Decimal]) -> None:
        self._debit_amount = value

    @property
    def credit_amount(self) -> Optional[Decimal]:
        """The credit amount of this accounting record."""
        if self._credit_amount is None:
            self._credit_amount = self.amount if self.is_credit else None
        return self._credit_amount

    @credit_amount.setter
    def credit_amount(self, value: Optional[Decimal]):
        self._credit_amount = value

    @property
    def is_balanced(self) -> bool:
        """Whether the transaction of this record is balanced. """
        if self._is_balanced is None:
            self._is_balanced = self.transaction.is_balanced
        return self._is_balanced

    @is_balanced.setter
    def is_balanced(self, value: bool) -> None:
        self._is_balanced = value

    @property
    def has_order_hole(self) -> bool:
        """Whether the order of the transactions on this day is not
        1, 2, 3, 4, 5..., and should be reordered. """
        if self._has_order_hole is None:
            self._has_order_hole = self.transaction.has_order_hole
        return self._has_order_hole

    @has_order_hole.setter
    def has_order_hole(self, value: bool) -> None:
        self._has_order_hole = value



# The core application of the Mia project.
#   by imacat <imacat@mail.imacat.idv.tw>, 2020/7/1

#  Copyright (c) 2020 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The utilities of the Mia core application.

"""
import datetime
import urllib.parse
from secrets import randbelow
from typing import Dict, List, Any, Type, Optional

from django.conf import settings
from django.db.models import Model
from django.http import HttpRequest
from django.urls import reverse
from django.utils.translation import pgettext, get_language


def new_pk(cls: Type[Model]) -> int:
    """Finds a random ID that does not conflict with the existing data records.

    Args:
        cls: The Django model class.

    Returns:
         The new random ID.
    """
    while True:
        pk = 100000000 + randbelow(900000000)
        try:
            cls.objects.get(pk=pk)
        except cls.DoesNotExist:
            return pk


def strip_post(post: Dict[str, str]) -> None:
    """Strips the values of the POSTed data.  Empty strings are removed.

    Args:
        post (dict[str]): The POSTed data.
    """
    for key in list(post.keys()):
        post[key] = post[key].strip()
        if post[key] == "":
            del post[key]


STORAGE_KEY: str = "stored_post"


def store_post(request: HttpRequest, post: Dict[str, str]):
    """Stores the POST data into the session.

    Args:
        request: The request.
        post: The POST data.
    """
    request.session[STORAGE_KEY] = post


def retrieve_store(request: HttpRequest) -> Optional[Dict[str, str]]:
    """Retrieves the POST data from the storage.

    Args:
        request: The request.

    Returns:
        The POST data, or None if the previously-stored data does not exist.
    """
    if STORAGE_KEY not in request.session:
        return None
    post = request.session[STORAGE_KEY]
    del request.session[STORAGE_KEY]
    return post


def parse_date(s: str):
    """Parses a string for a date.  The date can be either YYYY-MM-DD,
    Y/M/D, or M/D/Y.

    Args:
        s: The string.

    Returns:
        The date.

    Raises:
        ValueError: When the string is not in a valid format.
    """
    for f in ["%Y-%m-%d", "%m/%d/%Y", "%Y/%m/%d"]:
        try:
            return datetime.datetime.strptime(s, f)
        except ValueError:
            pass
    raise ValueError(F"not a recognized date {s}")


class Language:
    """A language.

    Args:
        language: The Django language code.

    Attributes:
        id (str): The language ID
        db (str): The database column suffix of this language.
        locale (str); The locale name of this language.
        is_default (bool): Whether this is the default language.
    """
    def __init__(self, language: str):
        self.id = language
        self.db = "_" + language.lower().replace("-", "_")
        if language == "zh-hant":
            self.locale = "zh-TW"
        elif language == "zh-hans":
            self.locale = "zh-CN"
        else:
            self.locale = language
        self.is_default = (language == settings.LANGUAGE_CODE)

    @staticmethod
    def default():
        return Language(settings.LANGUAGE_CODE)

    @staticmethod
    def current():
        return Language(get_language())


class UrlBuilder:
    """The URL builder.

    Attributes:
        path (str): the base path
        params (list[Param]): The query parameters
    """
    def __init__(self, start_url: str):
        """Constructs a new URL builder.

        Args:
            start_url (str): The URL to start with
        """
        pos = start_url.find("?")
        if pos == -1:
            self.path = start_url
            self.params = []
            return
        self.path = start_url[:pos]
        self.params = []
        for piece in start_url[pos + 1:].split("&"):
            pos = piece.find("=")
            name = urllib.parse.unquote_plus(piece[:pos])
            value = urllib.parse.unquote_plus(piece[pos + 1:])
            self.params.append(self.Param(name, value))

    def add(self, name, value):
        """Adds a query parameter.

        Args:
            name (str): The parameter name
            value (str): The parameter value

        Returns:
            UrlBuilder: The URL builder itself, with the parameter
                modified.
        """
        if value is not None:
            self.params.append(self.Param(name, value))
        return self

    def remove(self, name):
        """Removes a query parameter.

        Args:
            name (str): The parameter name

        Returns:
            UrlBuilder: The URL builder itself, with the parameter
                modified.
        """
        self.params = [x for x in self.params if x.name != name]
        return self

    def query(self, **kwargs):
        """A keyword-styled query parameter setter.  The existing values are
        always replaced.  Multiple-values are added when the value is a list or
        tuple.  The existing values are dropped when the value is None.
        """
        for key in kwargs:
            self.remove(key)
            if isinstance(kwargs[key], list) or isinstance(kwargs[key], tuple):
                for value in kwargs[key]:
                    self.add(key, value)
            elif kwargs[key] is None:
                pass
            else:
                self.add(key, kwargs[key])
        return self

    def clone(self):
        """Returns a copy of this URL builder.

        Returns:
            UrlBuilder: A copy of this URL builder.
        """
        another = UrlBuilder(self.path)
        another.params = [
            self.Param(x.name, x.value) for x in self.params]
        return another

    def __str__(self) -> str:
        if len(self.params) == 0:
            return self.path
        return self.path + "?" + "&".join([
            str(x) for x in self.params])

    class Param:
        """A query parameter.

        Attributes:
            name: The parameter name
            value: The parameter value
        """
        def __init__(self, name: str, value: str):
            """Constructs a new query parameter

            Args:
                name (str): The parameter name
                value (str): The parameter value
            """
            self.name = name
            self.value = value

        def __str__(self) -> str:
            """Returns the string representation of this query
            parameter.

            Returns:
                str: The string representation of this query
                    parameter
            """
            return "%s=%s" % (
                urllib.parse.quote(self.name),
                urllib.parse.quote(self.value))


class Pagination:
    """The pagination.

    Args:
        request: The request.
        items: All the items.
        is_reversed: Whether we should display the last page first.

    Raises:
        PaginationException: With invalid pagination parameters

    Attributes:
        current_url (UrlBuilder): The current request URL.
        is_reversed (bool): Whether we should display the last page first.
        page_size (int): The page size.
        total_pages (int): The total number of pages available.
        is_paged (bool): Whether there are more than one page.
        page_no (int): The current page number.
        items (list[Model]): The items in the current page.
    """
    DEFAULT_PAGE_SIZE = 10

    def __init__(self, request: HttpRequest, items: List[Any],
                 is_reversed: bool = False):
        self.current_url = UrlBuilder(request.get_full_path())
        self.is_reversed = is_reversed
        self.page_size = self.DEFAULT_PAGE_SIZE
        self.total_pages = None
        self.is_paged = None
        self.page_no = 1
        self.items = []

        # The page size
        try:
            self.page_size = int(request.GET["page-size"])
            if self.page_size == self.DEFAULT_PAGE_SIZE:
                raise PaginationException(self.current_url.remove("page-size"))
            if self.page_size < 1:
                raise PaginationException(self.current_url.remove("page-size"))
        except KeyError:
            self.page_size = self.DEFAULT_PAGE_SIZE
        except ValueError:
            raise PaginationException(self.current_url.remove("page-size"))
        self.total_pages = int(
            (len(items) - 1) / self.page_size) + 1
        default_page_no = 1 if not is_reversed else self.total_pages
        self.is_paged = self.total_pages > 1

        # The page number
        try:
            self.page_no = int(request.GET["page"])
            if not self.is_paged:
                raise PaginationException(self.current_url.remove("page"))
            if self.page_no == default_page_no:
                raise PaginationException(self.current_url.remove("page"))
            if self.page_no < 1:
                raise PaginationException(self.current_url.remove("page"))
            if self.page_no > self.total_pages:
                raise PaginationException(self.current_url.remove("page"))
        except KeyError:
            self.page_no = default_page_no
        except ValueError:
            raise PaginationException(self.current_url.remove("page"))

        if not self.is_paged:
            self.page_no = 1
            self.items = items
            return
        start_no = self.page_size * (self.page_no - 1)
        self.items = items[start_no:start_no + self.page_size]

    def links(self):
        """Returns the navigation links of the pagination bar.

        Returns:
            List[Link]: The navigation links of the pagination bar.
        """
        base_url = self.current_url.clone().remove("page").remove("s")
        links = []
        # The previous page
        link = self.Link()
        link.title = pgettext("Pagination|", "Previous")
        if self.page_no > 1:
            if self.page_no - 1 == 1:
                if not self.is_reversed:
                    link.url = str(base_url)
                else:
                    link.url = str(base_url.clone().add(
                        "page", "1"))
            else:
                link.url = str(base_url.clone().add(
                    "page", str(self.page_no - 1)))
        link.is_small_screen = True
        links.append(link)
        # The first page
        link = self.Link()
        link.title = "1"
        if not self.is_reversed:
            link.url = str(base_url)
        else:
            link.url = str(base_url.clone().add(
                "page", "1"))
        if self.page_no == 1:
            link.is_active = True
        links.append(link)
        # The previous ellipsis
        if self.page_no > 4:
            link = self.Link()
            if self.page_no > 5:
                link.title = pgettext("Pagination|", "...")
            else:
                link.title = "2"
                link.url = str(base_url.clone().add(
                    "page", "2"))
            links.append(link)
        # The nearby pages
        for no in range(self.page_no - 2, self.page_no + 3):
            if no <= 1 or no >= self.total_pages:
                continue
            link = self.Link()
            link.title = str(no)
            link.url = str(base_url.clone().add(
                "page", str(no)))
            if no == self.page_no:
                link.is_active = True
            links.append(link)
        # The next ellipsis
        if self.page_no + 3 < self.total_pages:
            link = self.Link()
            if self.page_no + 4 < self.total_pages:
                link.title = pgettext("Pagination|", "...")
            else:
                link.title = str(self.total_pages - 1)
                link.url = str(base_url.clone().add(
                    "page", str(self.total_pages - 1)))
            links.append(link)
        # The last page
        link = self.Link()
        link.title = str(self.total_pages)
        if self.is_reversed:
            link.url = str(base_url)
        else:
            link.url = str(base_url.clone().add(
                "page", str(self.total_pages)))
        if self.page_no == self.total_pages:
            link.is_active = True
        links.append(link)
        # The next page
        link = self.Link()
        link.title = pgettext("Pagination|", "Next")
        if self.page_no < self.total_pages:
            if self.page_no + 1 == self.total_pages:
                if self.is_reversed:
                    link.url = str(base_url)
                else:
                    link.url = str(base_url.clone().add(
                        "page", str(self.total_pages)))
            else:
                link.url = str(base_url.clone().add(
                    "page", str(self.page_no + 1)))
        link.is_small_screen = True
        links.append(link)
        return links

    class Link:
        """A navigation link in the pagination bar.

        Attributes:
            url (str): The link URL, or for a non-link slot.
            title (str): The title of the link.
            is_active (bool): Whether this link is currently active.
            is_small_screen (bool): Whether this link is for small screens.
        """
        def __int__(self):
            self.url = None
            self.title = None
            self.is_active = False
            self.is_small_screen = False

    def page_size_options(self):
        """Returns the page size options.

        Returns:
            List[PageSizeOption]: The page size options.
        """
        base_url = self.current_url.remove("page").remove("page-size")
        return [self.PageSizeOption(x, self._page_size_url(base_url, x))
                for x in [10, 100, 200]]

    @staticmethod
    def _page_size_url(base_url: UrlBuilder, size: int) -> str:
        """Returns the URL for a new page size.

        Args:
            base_url (UrlBuilder): The base URL builder.
            size (int): The new page size.

        Returns:
            str: The URL for the new page size.
        """
        if size == Pagination.DEFAULT_PAGE_SIZE:
            return str(base_url)
        return str(base_url.clone().add("page-size", str(size)))

    class PageSizeOption:
        """A page size option.

        Args:
            size: The page size.
            url: The URL of this page size.

        Attributes:
            size (int): The page size.
            url (str): The URL for this page size.
        """
        def __init__(self, size: int, url: str):
            self.size = size
            self.url = url


class PaginationException(Exception):
    """The exception thrown with invalid pagination parameters.

    Args:
        url_builder: The canonical URL to redirect to.

    Attributes:
        url (str): The canonical URL to redirect to.
    """
    def __init__(self, url_builder: UrlBuilder):
        self.url = str(url_builder)


CDN_LIBRARIES = {
    "jquery": {"css": [],
               "js": ["https://code.jquery.com/jquery-3.5.1.min.js"]},
    "bootstrap4": {
        "css": [("https://stackpath.bootstrapcdn.com/bootstrap/4.1.3/css/"
                 "bootstrap.min.css")],
        "js": [("https://cdnjs.cloudflare.com/ajax/libs/popper.js/1.14.3/umd/"
                "popper.min.js"),
               ("https://stackpath.bootstrapcdn.com/bootstrap/4.1.3/js/"
                "bootstrap.min.js")]},
    "font-awesome-5": {
        "css": ["https://use.fontawesome.com/releases/v5.14.0/css/all.css"],
        "js": []},
    "bootstrap4-datatables": {
        "css": [("https://cdn.datatables.net/1.10.21/css/"
                 "jquery.dataTables.min.css"),
                ("https://cdn.datatables.net/1.10.21/css/"
                 "dataTables.bootstrap4.min.css")],
        "js": [("https://cdn.datatables.net/1.10.21/js/"
                "jquery.dataTables.min.js"),
               ("https://cdn.datatables.net/1.10.21/js/"
                "dataTables.bootstrap4.min.js")]},
    "jquery-ui": {"css": [("https://cdnjs.cloudflare.com/ajax/libs/jqueryui/"
                           "1.12.1/jquery-ui.min.css")],
                  "js": [("https://cdnjs.cloudflare.com/ajax/libs/jqueryui/"
                          "1.12.1/jquery-ui.min.js")]},
    "bootstrap4-tempusdominus": {
        "css": [("https://cdnjs.cloudflare.com/ajax/libs/"
                 "tempusdominus-bootstrap-4/5.1.2/css/"
                 "tempusdominus-bootstrap-4.min.css")],
        "js": [("https://cdnjs.cloudflare.com/ajax/libs/moment.js/2.27.0/"
                "moment-with-locales.min.js"),
               ("https://cdnjs.cloudflare.com/ajax/libs/"
                "tempusdominus-bootstrap-4/5.1.2/js/"
                "tempusdominus-bootstrap-4.js")]},
    "decimal-js": {"css": [],
                   "js": [("https://cdnjs.cloudflare.com/ajax/libs/decimal.js/"
                           "10.2.0/decimal.min.js")]},
    "period-chooser": {"css": ["mia_core/css/period-chooser.css"],
                       "js": ["mia_core/js/period-chooser.js"]}
}
DEFAULT_LIBS = []


class CssAndJavaScriptLibraries:
    """The CSS and JavaScript library resolver."""
    AVAILABLE_LIBS: List[str] = ["jquery", "bootstrap4", "font-awesome-5",
                                 "bootstrap4-datatables", "jquery-ui",
                                 "bootstrap4-tempusdominus", "decimal-js",
                                 "i18n", "period-chooser"]

    def __init__(self, *args):
        self._use: Dict[str, bool] = {x: False for x in self.AVAILABLE_LIBS}
        self._add_default_libs()
        # The specified libraries
        if len(args) > 0:
            libs = args[0]
            invalid = [x for x in libs if x not in self.AVAILABLE_LIBS]
            if len(invalid) > 0:
                raise NameError("library %s invalid" % ", ".join(invalid))
            for lib in libs:
                self._use[lib] = True
        self._css = []
        try:
            self._css = self._css + settings.DEFAULT_CSS
        except AttributeError:
            pass
        self._js = []
        try:
            self._css = self._css + settings.DEFAULT_JS
        except AttributeError:
            pass

    def _add_default_libs(self):
        """Adds the default libraries."""
        invalid = [x for x in DEFAULT_LIBS if x not in self.AVAILABLE_LIBS]
        if len(invalid) > 0:
            raise NameError("library %s invalid" % ", ".join(invalid))
        for lib in DEFAULT_LIBS:
            self._use[lib] = True

    def use(self, *args) -> None:
        """Use the specific libraries.

        Args:
            args: The libraries.
        """
        if len(args) == 0:
            return
        libs = args[0]
        invalid = [x for x in libs if x not in self.AVAILABLE_LIBS]
        if len(invalid) > 0:
            raise NameError("library %s invalid" % ", ".join(invalid))
        for lib in libs:
            self._use[lib] = True

    def add_css(self, css) -> None:
        """Adds a custom CSS file."""
        self._css.append(css)

    def add_js(self, js) -> None:
        """Adds a custom JavaScript file."""
        self._js.append(js)

    def css(self) -> List[str]:
        """Returns the stylesheet files to use."""
        use: Dict[str, bool] = self._solve_use_dependencies()
        css = []
        for lib in [x for x in self.AVAILABLE_LIBS if use[x]]:
            if lib == "i18n":
                continue
            try:
                css = css + settings.STATIC_LIBS[lib]["css"]
            except AttributeError:
                css = css + CDN_LIBRARIES[lib]["css"]
            except TypeError:
                css = css + CDN_LIBRARIES[lib]["css"]
            except KeyError:
                css = css + CDN_LIBRARIES[lib]["css"]
        return css + self._css

    def js(self) -> List[str]:
        """Returns the JavaScript files to use."""
        use: Dict[str, bool] = self._solve_use_dependencies()
        js = []
        for lib in [x for x in self.AVAILABLE_LIBS if use[x]]:
            if lib == "i18n":
                js.append(reverse("javascript-catalog"))
                continue
            try:
                js = js + settings.STATIC_LIBS[lib]["js"]
            except AttributeError:
                js = js + CDN_LIBRARIES[lib]["js"]
            except TypeError:
                js = js + CDN_LIBRARIES[lib]["js"]
            except KeyError:
                js = js + CDN_LIBRARIES[lib]["js"]
        return js + self._js

    def _solve_use_dependencies(self) -> Dict[str, bool]:
        """Solves and returns the library dependencies."""
        use: Dict[str, bool] = {x: self._use[x] for x in self._use}
        if use["period-chooser"]:
            use["bootstrap4-tempusdominus"] = True
        if use["bootstrap4-tempusdominus"]:
            use["bootstrap4"] = True
        if use["bootstrap4-datatables"]:
            use["bootstrap4"] = True
        if use["jquery-ui"]:
            use["jquery"] = True
        if use["bootstrap4"]:
            use["jquery"] = True
        return use


def add_default_libs(*args) -> None:
    """Adds the specified libraries to the default CSS and JavaScript
    libraries.

    Args:
        args: The libraries to be added to the default libraries
    """
    libs = args
    invalid = [x for x in libs
               if x not in CssAndJavaScriptLibraries.AVAILABLE_LIBS]
    if len(invalid) > 0:
        raise NameError("library %s invalid" % ", ".join(invalid))
    for lib in libs:
        if lib not in DEFAULT_LIBS:
            DEFAULT_LIBS.append(lib)



# The accounting application of the Mia project.
#   by imacat <imacat@mail.imacat.idv.tw>, 2020/7/23

#  Copyright (c) 2020 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The URL converters.

"""
import datetime
import re

from django.utils.translation import gettext as _

from .models import Transaction, Record, Account
from mia_core.period import Period


class TransactionTypeConverter:
    """The path converter for the transaction types."""
    regex = "income|expense|transfer"

    def to_python(self, value):
        return value

    def to_url(self, value):
        return value


class PeriodConverter:
    """The path converter for the period."""
    regex = (r"([0-9]{4}(-[0-9]{2}(-[0-9]{2})?)?)|"
             r"([0-9]{4}(-[0-9]{2}(-[0-9]{2})?)?)?-"
             r"([0-9]{4}(-[0-9]{2}(-[0-9]{2})?)?)?")

    def to_python(self, value):
        """Returns the period by the period specification.

        Args:
            value (str): The period specification.

        Returns:
            Period: The period.

        Raises:
            ValueError: When the period specification is invalid.
        """
        first_txn = Transaction.objects.order_by("date").first()
        data_start = first_txn.date if first_txn is not None else None
        last_txn = Transaction.objects.order_by("-date").first()
        data_end = last_txn.date if last_txn is not None else None
        # Raises ValueError
        return Period(value, data_start, data_end)

    def to_url(self, value):
        """Returns the specification of a period.

        Args:
            value (Period|str): The period.

        Returns:
            str: The period specification.
        """
        if isinstance(value, Period):
            return value.spec
        return value


class DateConverter:
    """The path converter for the date."""
    regex = "([0-9]{4})-([0-9]{2})-([0-9]{2})"

    def to_python(self, value):
        """Returns the date by the date specification.

        Args:
            value (str): The date specification.

        Returns:
            datetime.date: The date.
        """
        m = re.match(r"^(\d{4})-(\d{2})-(\d{2})$", value)
        year = int(m.group(1))
        month = int(m.group(2))
        day = int(m.group(3))
        return datetime.date(year, month, day)

    def to_url(self, value):
        """Returns the specification of a date.

        Args:
            value (datetime.date): The date.

        Returns:
            str: The date specification.
        """
        return value.strftime("%Y-%m-%d")


class AccountConverter:
    """The path converter for the account."""
    regex = "[1-9]{1,5}"

    def to_python(self, value):
        """Returns the account by the account code.

        Args:
            value (str): The account code.

        Returns:
            Account: The account.
        """
        try:
            return Account.objects.get(code=value)
        except Account.DoesNotExist:
            raise ValueError

    def to_url(self, value):
        """Returns the code of an account.

        Args:
            value (Account): The account.

        Returns:
            str: The account code.
        """
        return value.code


class CashAccountConverter:
    """The path converter for the cash account."""
    regex = "0|(11|12|21|22)[1-9]{1,3}"

    def to_python(self, value: str) -> Account:
        """Returns the cash account by the account code.

        Args:
            value: The account code.

        Returns:
            The account.

        Raises:
            ValueError: When the value is invalid
        """
        if value == "0":
            return Account(
                code="0",
                title=_("current assets and liabilities"),
            )
        try:
            account = Account.objects.get(code=value)
        except Account.DoesNotExist:
            raise ValueError
        if Record.objects.filter(account=account).count() == 0:
            raise ValueError
        return account

    def to_url(self, value: Account) -> str:
        """Returns the code of an account.

        Args:
            value: The account.

        Returns:
            The account code.
        """
        return value.code


class LedgerAccountConverter:
    """The path converter for the ledger account."""
    regex = "[1-9]{1,5}"

    def to_python(self, value: str) -> Account:
        """Returns the ledger account by the account code.

        Args:
            value: The account code.

        Returns:
            The account.

        Raises:
            ValueError: When the value is invalid
        """
        try:
            account = Account.objects.get(code=value)
        except Account.DoesNotExist:
            raise ValueError
        if Record.objects.filter(account__code__startswith=value).count() == 0:
            raise ValueError
        return account

    def to_url(self, value: Account) -> str:
        """Returns the code of an account.

        Args:
            value: The account.

        Returns:
            The account code.
        """
        return value.code


class TransactionConverter:
    """The path converter for the accounting transactions."""
    regex = "[1-9][0-9]{8}"

    def to_python(self, value: str) -> Transaction:
        """Returns the transaction by the transaction ID.

        Args:
            value: The transaction ID.

        Returns:
            The account.

        Raises:
            ValueError: When the value is invalid
        """
        try:
            return Transaction.objects.get(pk=value)
        except Transaction.DoesNotExist:
            raise ValueError

    def to_url(self, value: Transaction) -> str:
        """Returns the ID of a transaction.

        Args:
            value: The transaction.

        Returns:
            The transaction ID.
        """
        return value.pk



# The core application of the Mia project.
#   by imacat <imacat@mail.imacat.idv.tw>, 2020/7/4

#  Copyright (c) 2020 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The views of the Mia core application.

"""
from typing import Dict, Type, Optional, Any

from dirtyfields import DirtyFieldsMixin
from django import forms
from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.db import transaction
from django.db.models import Model
from django.http import HttpResponse, HttpRequest, \
    HttpResponseRedirect, Http404
from django.shortcuts import redirect, render
from django.urls import reverse
from django.views.generic import DeleteView as CoreDeleteView, \
    RedirectView as CoreRedirectView
from django.views.generic.base import View

from . import utils
from .models import StampedModel
from .utils import UrlBuilder


class RedirectView(CoreRedirectView):
    """The redirect view, with current_app at the current namespace."""

    def get_redirect_url(self, *args, **kwargs):
        url = reverse(self.pattern_name, kwargs=kwargs,
                      current_app=self.request.resolver_match.namespace)
        if self.query_string and self.request.META["QUERY_STRING"] != "":
            url = url + "?" + self.request.META["QUERY_STRING"]
        return url


class FormView(View):
    """The base form view."""
    model: Type[Model] = None
    form_class: Type[forms.Form] = None
    template_name: str = None
    context_object_name: str = "form"
    success_url: str = None
    error_url: str = None
    not_modified_message: str = None
    success_message: str = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.object: Optional[Model] = None

    def dispatch(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        """The view to store an accounting transaction."""
        self.object = self.get_object()
        if self.request.method == "POST":
            return self.post(request, *args, **kwargs)
        else:
            return self.get(request, *args, **kwargs)

    def get(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        """Handles the GET requests."""
        return render(self.request, self.get_template_name(),
                      self.get_context_data(**kwargs))

    def post(self, request: HttpRequest, *args,
             **kwargs) -> HttpResponseRedirect:
        """Handles the POST requests."""
        form = self.get_form(**kwargs)
        if not form.is_valid():
            return self.form_invalid(form)
        return self.form_valid(form)

    def get_form_class(self) -> Type[forms.Form]:
        """Returns the form class."""
        if self.form_class is None:
            raise AttributeError("Please defined the form_class property.")
        return self.form_class

    @property
    def _model(self):
        if self.model is None:
            raise AttributeError("Please defined the model property.")
        return self.model

    def get_context_data(self, **kwargs) -> Dict[str, Any]:
        """Returns the context data for the template."""
        return {self.context_object_name: self.get_form()}

    def get_form(self, **kwargs) -> forms.Form:
        """Returns the form for the template."""
        if self.request.method == "POST":
            post = self.request.POST.dict()
            utils.strip_post(post)
            return self.make_form_from_post(post)
        else:
            previous_post = utils.retrieve_store(self.request)
            if previous_post is not None:
                return self.make_form_from_post(previous_post)
            if self.object is not None:
                return self.make_form_from_model(self.object)
            return self.get_form_class()()

    def get_template_name(self) -> str:
        """Returns the name of the template."""
        if self.template_name is not None:
            return self.template_name
        if self.model is not None:
            app_name = self.request.resolver_match.app_name
            model_name = self.model.__name__.lower()
            return F"{app_name}/{model_name}_form.html"
        raise AttributeError(
            "Please either define the template_name or the model property.")

    def make_form_from_post(self, post: Dict[str, str]) -> forms.Form:
        """Creates and returns the form from the POST data."""
        return self.get_form_class()(post)

    def make_form_from_model(self, obj: Model) -> forms.Form:
        """Creates and returns the form from a data model."""
        form_class = self.get_form_class()
        return form_class({x: getattr(obj, x, None)
                           for x in form_class.base_fields})

    def fill_model_from_form(self, obj: Model, form: forms.Form) -> None:
        """Fills in the data model from the form."""
        for name in form.fields:
            setattr(obj, name, form[name].value())
        if isinstance(obj, StampedModel):
            obj.current_user = self.request.user

    def form_invalid(self, form: forms.Form) -> HttpResponseRedirect:
        """Handles the action when the POST form is invalid."""
        utils.store_post(self.request, form.data)
        return redirect(self.get_error_url())

    def form_valid(self, form: forms.Form) -> HttpResponseRedirect:
        """Handles the action when the POST form is valid."""
        if self.object is None:
            self.object = self._model()
        self.fill_model_from_form(self.object, form)
        if isinstance(self.object, DirtyFieldsMixin)\
                and not self.object.is_dirty(check_relationship=True):
            message = self.get_not_modified_message(form.cleaned_data)
        else:
            with transaction.atomic():
                self.object.save()
            message = self.get_success_message(form.cleaned_data)
        messages.success(self.request, message)
        return redirect(str(UrlBuilder(self.get_success_url())
                            .query(r=self.request.GET.get("r"))))

    def get_success_url(self) -> str:
        """Returns the URL on success."""
        if self.success_url is not None:
            return self.success_url
        get_absolute_url = getattr(self.object, "get_absolute_url", None)
        if get_absolute_url is not None:
            return get_absolute_url()
        raise AttributeError(
            "Please define either the success_url property,"
            " the get_absolute_url method on the data model,"
            " or the get_success_url method.")

    def get_error_url(self) -> str:
        """Returns the URL on error"""
        if self.error_url is not None:
            return self.error_url
        return self.request.get_full_path()

    def get_not_modified_message(self, cleaned_data: Dict[str, str]) -> str:
        """Returns the message when the data was not modified.

        Args:
            cleaned_data: The cleaned data of the form.

        Returns:
            The message when the data was not modified.
        """
        return self.not_modified_message % cleaned_data

    def get_success_message(self, cleaned_data: Dict[str, str]) -> str:
        """Returns the success message.

        Args:
            cleaned_data: The cleaned data of the form.

        Returns:
            The message when the data was not modified.
        """
        return self.success_message % cleaned_data

    def get_object(self) -> Optional[Model]:
        """Finds and returns the current object, or None on a create form."""
        if "pk" in self.kwargs:
            pk = self.kwargs["pk"]
            try:
                return self._model.objects.get(pk=pk)
            except self._model.DoesNotExist:
                raise Http404
        return None


class DeleteView(SuccessMessageMixin, CoreDeleteView):
    """The delete form view, with SuccessMessageMixin."""

    def delete(self, request, *args, **kwargs):
        response = super(DeleteView, self).delete(request, *args, **kwargs)
        messages.success(request, self.get_success_message({}))
        return response



# The core application of the Mia project.
#   by imacat <imacat@mail.imacat.idv.tw>, 2020/7/1

#  Copyright (c) 2020 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The template tags and filters of the Mia core application.

"""
import datetime
import re
from datetime import date
from typing import Any

import titlecase
from django import template
from django.http import HttpRequest
from django.template import defaultfilters, RequestContext
from django.urls import reverse
from django.utils import timezone
from django.utils.safestring import SafeString
from django.utils.translation import gettext, get_language

from mia_core.utils import UrlBuilder, CssAndJavaScriptLibraries

register = template.Library()


@register.simple_tag(takes_context=True)
def setvar(context: RequestContext, key: str, value: Any) -> str:
    """Sets a variable in the template.

    Args:
        context: the context
        key: The variable name
        value: The variable value

    Returns:
        An empty string.
    """
    context.dicts[0][key] = value
    return ""


@register.simple_tag(takes_context=True)
def url_period(context: RequestContext, period_spec: str) -> str:
    """Returns the current URL with a new period.

    Args:
        context: The request context.
        period_spec: The period specification.

    Returns:
        The current URL with the new period.
    """
    view_name = "%s:%s" % (
        context.request.resolver_match.app_name,
        context.request.resolver_match.url_name)
    kwargs = context.request.resolver_match.kwargs.copy()
    kwargs["period"] = period_spec
    namespace = context.request.resolver_match.namespace
    return reverse(view_name, kwargs=kwargs, current_app=namespace)


@register.simple_tag(takes_context=True)
def url_with_return(context: RequestContext, url: str) -> str:
    """Returns the URL with the current page added as the "r" query parameter,
    so that returning to this page is possible.

    Args:
        context: The request context.
        url: The URL.

    Returns:
        The URL with the current page added as the "r" query parameter.
    """
    return str(UrlBuilder(url).query(
        r=str(UrlBuilder(context.request.get_full_path()).remove("s"))))


@register.simple_tag(takes_context=True)
def url_keep_return(context: RequestContext, url: str) -> str:
    """Returns the URL with the current "r" query parameter set, so that the
    next processor can still return to the same page.

    Args:
        context: The request context.
        url: The URL.

    Returns:
        The URL with the current "r" query parameter set.
    """
    return str(UrlBuilder(url).query(r=context.request.GET.get("r")))


@register.simple_tag(takes_context=True)
def init_libs(context: RequestContext) -> str:
    """Initializes the static libraries.

    Args:
        context: The request context.

    Returns:
        An empty string.
    """
    if "libs" not in context.dicts[0]:
        context.dicts[0]["libs"] = CssAndJavaScriptLibraries()
    return ""


@register.simple_tag(takes_context=True)
def add_lib(context: RequestContext, *args) -> str:
    """Adds CSS and JavaScript libraries.

    Args:
        context: The request context.
        args: The names of the CSS and JavaScript libraries.

    Returns:
        An empty string.
    """
    if "libs" not in context.dicts[0]:
        context.dicts[0]["libs"] = CssAndJavaScriptLibraries(args)
    else:
        context.dicts[0]["libs"].use(args)
    return ""


@register.simple_tag(takes_context=True)
def add_css(context: RequestContext, url: str) -> str:
    """Adds a local CSS file.  The file is added to the "css" template
    list variable.

    Args:
        context: The request context.
        url: The URL or path of the CSS file.

    Returns:
        An empty string.
    """
    if "libs" not in context.dicts[0]:
        context.dicts[0]["libs"] = CssAndJavaScriptLibraries()
    context.dicts[0]["libs"].add_css(url)
    return ""


@register.simple_tag(takes_context=True)
def add_js(context: RequestContext, url: str) -> str:
    """Adds a local JavaScript file.  The file is added to the "js" template
    list variable.

    Args:
        context: The request context.
        url: The URL or path of the JavaScript file.

    Returns:
        An empty string.
    """
    if "libs" not in context.dicts[0]:
        context.dicts[0]["libs"] = CssAndJavaScriptLibraries()
    context.dicts[0]["libs"].add_js(url)
    return ""


@register.filter
def smart_date(value: datetime.date) -> str:
    """Formats the date for human friendliness.

    Args:
        value: The date.

    Returns:
        The human-friendly format of the date.
    """
    if value == date.today():
        return gettext("Today")
    prev_days = (value - date.today()).days
    if prev_days == -1:
        return gettext("Yesterday")
    if prev_days == 1:
        return gettext("Tomorrow")
    if get_language() == "zh-hant":
        if prev_days == -2:
            return "前天"
        if prev_days == -3:
            return "大前天"
        if prev_days == 2:
            return "後天"
        if prev_days == 3:
            return "大後天"
    if date.today().year == value.year:
        return defaultfilters.date(value, "n/j(D)").replace("星期", "")
    return defaultfilters.date(value, "Y/n/j(D)").replace("星期", "")


@register.filter
def smart_month(value: datetime.date) -> str:
    """Formats the month for human friendliness.

    Args:
        value: The month.

    Returns:
        The human-friendly format of the month.
    """
    today = timezone.localdate()
    if value.year == today.year and value.month == today.month:
        return gettext("This Month")
    month = today.month - 1
    year = today.year
    if month < 1:
        month = 12
        year = year - 1
    if value.year == year and value.month == month:
        return gettext("Last Month")
    return defaultfilters.date(value, "Y/n")


@register.filter
def title_case(value: str) -> str:
    """Formats the title in a proper American-English case.

    Args:
        value: The title.

    Returns:
        The title in a proper American-English case.
    """
    value = str(value)
    if isinstance(value, SafeString):
        value = value + ""
    return titlecase.titlecase(value)


@register.filter
def is_in_section(request: HttpRequest, section_name: str) -> bool:
    """Returns whether the request is currently in a section.

    Args:
        request: The request.
        section_name: The view name of this section.

    Returns:
        True if the request is currently in this section, or False otherwise.
    """
    if request is None:
        return False
    if request.resolver_match is None:
        return False
    view_name = request.resolver_match.view_name
    return view_name == section_name\
        or view_name.startswith(section_name + ".")


@register.filter
def is_static_url(target: str) -> bool:
    """Returns whether the target URL is a static path

    Args:
        target: The target, either a static path that need to be passed to
            the static template tag, or an HTTP, HTTPS URL or absolute path
            that should be displayed directly.

    Returns:
        True if the target URL is a static path, or False otherwise.
    """
    return not (re.match("^https?://", target) or target.startswith("/"))



# The core application of the Mia project.
#   by imacat <imacat@mail.imacat.idv.tw>, 2020/8/1

#  Copyright (c) 2020 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The validators of the Mia core application.

"""
from django.core.exceptions import ValidationError
from django.db.models import Q
from django.utils.translation import gettext as _

from .models import Account, Record


def validate_record_id(value: str) -> None:
    """Validates the record ID.

    Args:
        value: The record ID.

    Raises:
        ValidationError: When the validation fails.
    """
    try:
        Record.objects.get(pk=value)
    except Record.DoesNotExist:
        raise ValidationError(_("This accounting record does not exists."),
                              code="not_exist")


def validate_record_account_code(value: str) -> None:
    """Validates an account code.

    Args:
        value: The account code.

    Raises:
        ValidationError: When the validation fails.
    """
    try:
        Account.objects.get(code=value)
    except Account.DoesNotExist:
        raise ValidationError(_("This account does not exist."),
                              code="not_exist")
    child = Account.objects.filter(
        Q(code__startswith=value),
        ~Q(code=value),
    ).first()
    if child is not None:
        raise ValidationError(_("You cannot select a parent account."),
                              code="parent_account")



# The accounting application of the Mia project.
#   by imacat <imacat@mail.imacat.idv.tw>, 2022/12/5

#  Copyright (c) 2022 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The tests to run the accounting application.

"""
import os
import sys
import unittest
from datetime import datetime
from pathlib import Path
from secrets import token_urlsafe

from django.core.management import call_command
from django.core.wsgi import get_wsgi_application
from django.test import Client


class RunTestCase(unittest.TestCase):
    """The test case to run the accounting application."""

    def setUp(self):
        """Sets up the test.

        Returns:
            None.
        """
        test_site_path: str = str(Path(__file__).parent / "test_site")
        if test_site_path not in sys.path:
            sys.path.append(test_site_path)
        self.username: str = "admin"
        self.password: str = token_urlsafe(16)
        os.environ.setdefault("DJANGO_SETTINGS_MODULE", "test_site.settings")
        get_wsgi_application()
        call_command("migrate")
        os.getlogin()
        call_command("createsuperuser", interactive=False,
                     username=self.username,  email="test@example.com")
        from django.contrib.auth.models import User
        user = User.objects.get(username=self.username)
        user.set_password(self.password)
        user.save()
        call_command("accounting_accounts")
        call_command("accounting_sample")
        self.client = Client()

    def test_run(self):
        """Tests the accounting application.

        Returns:
            None.
        """
        response = self.client.get("/accounting/", follow=True)
        self.assertEqual(response.status_code, 404)
        response = self.client.post("/admin/login/",
                                    {"username": self.username,
                                     "password": self.password,
                                     "next": "/ok"})
        # 200 for errors, 302 for success.
        self.assertEqual(response.status_code, 302)
        response = self.client.get("/accounting/", follow=True)
        self.assertEqual(response.status_code, 200)

        today: str = datetime.today().strftime("%Y-%m-%d")
        expense_add_uri: str = "/accounting/transactions/expense/create?r=/ok"
        income_add_uri: str = "/accounting/transactions/income/create?r=/ok"
        transfer_add_uri: str \
            = "/accounting/transactions/transfer/create?r=/ok"

        response = self.client.post(
            expense_add_uri,
            {"date": today,
             "debit-2-ord": 6,
             "debit-2-summary": "lunch",
             "debit-2-amount": 80,
             "debit-2-account": "6272",
             "debit-8-ord": 4,
             "debit-8-summary": "movies",
             "debit-8-amount": 320,
             "debit-8-account": "6273",
             "notes": "yammy"})
        self.assertEqual(response.status_code, 302)
        self.assertNotEqual(response.headers["Location"], expense_add_uri)

        response = self.client.post(
            income_add_uri,
            {"date": today,
             "credit-3-ord": 7,
             "credit-3-summary": "withdrawal",
             "credit-3-amount": 1000,
             "credit-3-account": "1113",
             "credit-6-ord": 3,
             "credit-6-summary": "payroll",
             "credit-6-amount": 10000,
             "credit-6-account": "4611",
             "notes": "wonderful"})
        self.assertEqual(response.status_code, 302)
        self.assertNotEqual(response.headers["Location"], income_add_uri)

        response = self.client.post(
            transfer_add_uri,
            {"date": today,
             "debit-2-ord": 6,
             "debit-2-summary": "lunch",
             "debit-2-amount": 80,
             "debit-2-account": "6272",
             "debit-8-ord": 4,
             "debit-8-summary": "movies",
             "debit-8-amount": 320,
             "debit-8-account": "6273",
             "credit-3-ord": 7,
             "credit-3-summary": "withdrawal",
             "credit-3-amount": 100,
             "credit-3-account": "1113",
             "credit-6-ord": 3,
             "credit-6-summary": "",
             "credit-6-amount": 320,
             "credit-6-account": "1111",
             "notes": "nothing"})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.headers["Location"], transfer_add_uri)

        response = self.client.post(
            transfer_add_uri,
            {"date": today,
             "debit-2-ord": 6,
             "debit-2-summary": "lunch",
             "debit-2-amount": 80,
             "debit-2-account": "6272",
             "debit-8-ord": 4,
             "debit-8-summary": "movies",
             "debit-8-amount": 320,
             "debit-8-account": "6273",
             "credit-3-ord": 7,
             "credit-3-summary": "withdrawal",
             "credit-3-amount": 100,
             "credit-3-account": "1113",
             "credit-6-ord": 3,
             "credit-6-summary": "",
             "credit-6-amount": 300,
             "credit-6-account": "1111",
             "notes": "nothing"})
        self.assertEqual(response.status_code, 302)
        self.assertNotEqual(response.headers["Location"], transfer_add_uri)



# The accounting application of the Mia project.
#   by imacat <imacat@mail.imacat.idv.tw>, 2020/7/13

#  Copyright (c) 2020 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The utilities of the accounting application.

"""
import datetime
from typing import Union, Tuple, List, Optional, Iterable

from django.conf import settings
from django.db.models import Q, Sum, Case, When, F, Count, Max, Min
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import gettext as _

from mia_core.period import Period
from mia_core.templatetags.mia_core import smart_month
from mia_core.utils import new_pk
from .models import Account, Transaction, Record

AccountData = Tuple[Union[str, int], str, str, str]
RecordData = Tuple[Union[str, int], Optional[str], float]

DEFAULT_CASH_ACCOUNT = "1111"
CASH_SHORTCUT_ACCOUNTS = ["0", "1111"]
DEFAULT_LEDGER_ACCOUNT = "1111"
PAYABLE_ACCOUNTS = ["2141", "21413"]
EQUIPMENT_ACCOUNTS = ["1441"],


class MonthlySummary:
    """A summary record.

    Args:
        month: The month.
        label: The text label.
        credit: The credit amount.
        debit: The debit amount.
        balance: The balance.
        cumulative_balance: The cumulative balance.

    Attributes:
        month (datetime.date): The month.
        label (str): The text label.
        credit (int): The credit amount.
        debit (int): The debit amount.
        balance (int): The balance.
        cumulative_balance (int): The cumulative balance.
    """

    def __init__(self, month: datetime.date = None, label: str = None,
                 credit: int = None, debit: int = None, balance: int = None,
                 cumulative_balance: int = None):
        self.month = month
        self.label = label
        self.credit = credit
        self.debit = debit
        self.balance = balance
        self.cumulative_balance = cumulative_balance
        if self.label is None and self.month is not None:
            self.label = smart_month(self.month)


class ReportUrl:
    """The URL of the accounting reports.

    Args:
        namespace: The namespace of the current application.
        cash: The currently-specified account of the cash account or cash
            summary.
        ledger: The currently-specified account of the ledger or leger summary.
        period: The currently-specified period.
    """

    def __init__(self, namespace: str, cash: Account = None,
                 ledger: Account = None, period: Period = None,):
        self._period = Period() if period is None else period
        self._cash = get_default_cash_account() if cash is None else cash
        self._ledger = get_default_ledger_account()\
            if ledger is None else ledger
        self._namespace = namespace

    def cash(self) -> str:
        return reverse("accounting:cash", args=[self._cash, self._period],
                       current_app=self._namespace)

    def cash_summary(self) -> str:
        return reverse("accounting:cash-summary", args=[self._cash],
                       current_app=self._namespace)

    def ledger(self) -> str:
        return reverse("accounting:ledger", args=[self._ledger, self._period],
                       current_app=self._namespace)

    def ledger_summary(self) -> str:
        return reverse("accounting:ledger-summary", args=[self._ledger],
                       current_app=self._namespace)

    def journal(self) -> str:
        return reverse("accounting:journal", args=[self._period],
                       current_app=self._namespace)

    def trial_balance(self) -> str:
        return reverse("accounting:trial-balance", args=[self._period],
                       current_app=self._namespace)

    def income_statement(self) -> str:
        return reverse("accounting:income-statement", args=[self._period],
                       current_app=self._namespace)

    def balance_sheet(self) -> str:
        return reverse("accounting:balance-sheet", args=[self._period],
                       current_app=self._namespace)


class DataFiller:
    """The helper to populate the accounting data.

    Args:
        user: The user in action.

    Attributes:
        user (User): The user in action.
    """

    def __init__(self, user):
        self.user = user

    def add_accounts(self, accounts: List[AccountData]) -> None:
        """Adds accounts.

        Args:
            accounts: Tuples of (code, English, Traditional Chinese, Simplified
                Chinese) of the accounts.
        """
        for data in accounts:
            code = data[0]
            if isinstance(code, int):
                code = str(code)
            parent = None if len(code) == 1\
                else Account.objects.get(code=code[:-1])
            account = Account(parent=parent, code=code, current_user=self.user)
            account.set_l10n_in("title", "en", data[1])
            account.set_l10n_in("title", "zh-hant", data[2])
            account.set_l10n_in("title", "zh-hans", data[3])
            account.save()

    def add_transfer_transaction(self, date: Union[datetime.date, int],
                                 debit: List[RecordData],
                                 credit: List[RecordData]) -> None:
        """Adds a transfer transaction.

        Args:
            date: The date, or the number of days from today.
            debit: Tuples of (account, summary, amount) of the debit records.
            credit: Tuples of (account, summary, amount) of the credit records.
        """
        if isinstance(date, int):
            date = timezone.localdate() + timezone.timedelta(days=date)
        order = Transaction.objects.filter(date=date).count() + 1
        transaction = Transaction(pk=new_pk(Transaction), date=date, ord=order,
                                  current_user=self.user)
        transaction.save()
        order = 1
        for data in debit:
            account = data[0]
            if isinstance(account, str):
                account = Account.objects.get(code=account)
            elif isinstance(account, int):
                account = Account.objects.get(code=str(account))
            transaction.record_set.create(pk=new_pk(Record), is_credit=False,
                                          ord=order, account=account,
                                          summary=data[1], amount=data[2],
                                          current_user=self.user)
            order = order + 1
        order = 1
        for data in credit:
            account = data[0]
            if isinstance(account, str):
                account = Account.objects.get(code=account)
            elif isinstance(account, int):
                account = Account.objects.get(code=str(account))
            transaction.record_set.create(pk=new_pk(Record), is_credit=True,
                                          ord=order, account=account,
                                          summary=data[1], amount=data[2],
                                          current_user=self.user)
            order = order + 1

    def add_income_transaction(self, date: Union[datetime.date, int],
                               credit: List[RecordData]) -> None:
        """Adds a cash income transaction.

        Args:
            date: The date, or the number of days from today.
            credit: Tuples of (account, summary, amount) of the credit records.
        """
        amount = sum([x[2] for x in credit])
        self.add_transfer_transaction(
            date, [(Account.CASH, None, amount)], credit)

    def add_expense_transaction(self, date: Union[datetime.date, int],
                                debit: List[RecordData]) -> None:
        """Adds a cash income transaction.

        Args:
            date: The date, or the number of days from today.
            debit: Tuples of (account, summary, amount) of the debit records.
        """
        amount = sum([x[2] for x in debit])
        self.add_transfer_transaction(
            date, debit, [(Account.CASH, None, amount)])


def get_cash_accounts() -> List[Account]:
    """Returns the cash accounts.

    Returns:
        The cash accounts.
    """
    accounts = list(
        Account.objects
        .filter(
            code__in=Record.objects
            .filter(
                Q(account__code__startswith="11")
                | Q(account__code__startswith="12")
                | Q(account__code__startswith="21")
                | Q(account__code__startswith="22"))
            .values("account__code"))
        .order_by("code"))
    accounts.insert(0, Account(
        code="0",
        title=_("current assets and liabilities"),
    ))
    return accounts


def get_default_cash_account() -> Account:
    """Returns the default cash account.

    Returns:
        The default cash account.
    """
    try:
        code = settings.ACCOUNTING["DEFAULT_CASH_ACCOUNT"]
    except AttributeError:
        code = DEFAULT_CASH_ACCOUNT
    except TypeError:
        code = DEFAULT_CASH_ACCOUNT
    except KeyError:
        code = DEFAULT_CASH_ACCOUNT
    if code == "0":
        return Account(code="0", title=_("current assets and liabilities"))
    try:
        return Account.objects.get(code=code)
    except Account.DoesNotExist:
        pass
    try:
        return Account.objects.get(code=DEFAULT_CASH_ACCOUNT)
    except Account.DoesNotExist:
        pass
    return Account(code="0", title=_("current assets and liabilities"))


def get_cash_shortcut_accounts() -> List[str]:
    """Returns the codes of the shortcut cash accounts.

    Returns:
        The codes of the shortcut cash accounts.
    """
    try:
        accounts = settings.ACCOUNTING["CASH_SHORTCUT_ACCOUNTS"]
    except AttributeError:
        return CASH_SHORTCUT_ACCOUNTS
    except TypeError:
        return CASH_SHORTCUT_ACCOUNTS
    except KeyError:
        return CASH_SHORTCUT_ACCOUNTS
    if not isinstance(accounts, list):
        return CASH_SHORTCUT_ACCOUNTS
    return accounts


def get_ledger_accounts() -> List[Account]:
    """Returns the accounts for the ledger.

    Returns:
        The accounts for the ledger.
    """
    """
    For SQL one-liner:
SELECT s.*
  FROM accounting_accounts AS s
  WHERE s.code IN (SELECT s.code
    FROM accounting_accounts AS s
      INNER JOIN (SELECT s.code
        FROM accounting_accounts AS s
         INNER JOIN accounting_records AS r ON r.account_id = s.id
        GROUP BY s.code) AS u
      ON u.code LIKE s.code || '%%'
    GROUP BY s.code)
  ORDER BY s.code
    """
    codes = {}
    for code in [x.code for x in Account.objects
                 .annotate(Count("record"))
                 .filter(record__count__gt=0)]:
        while len(code) > 0:
            codes[code] = True
            code = code[:-1]
    return Account.objects.filter(code__in=codes).order_by("code")


def get_default_ledger_account() -> Optional[Account]:
    """Returns the default ledger account.

    Returns:
        The default ledger account.
    """
    try:
        code = settings.ACCOUNTING["DEFAULT_CASH_ACCOUNT"]
    except AttributeError:
        code = DEFAULT_CASH_ACCOUNT
    except TypeError:
        code = DEFAULT_CASH_ACCOUNT
    except KeyError:
        code = DEFAULT_CASH_ACCOUNT
    try:
        return Account.objects.get(code=code)
    except Account.DoesNotExist:
        pass
    try:
        return Account.objects.get(code=DEFAULT_LEDGER_ACCOUNT)
    except Account.DoesNotExist:
        pass
    return None


def find_imbalanced(records: Iterable[Record]) -> None:
    """"Finds the records with imbalanced transactions, and sets their
    is_balanced attribute.

    Args:
        records: The accounting records.
    """
    imbalanced = [x.pk for x in Transaction.objects
                  .annotate(
                    balance=Sum(Case(
                        When(record__is_credit=True, then=-1),
                        default=1) * F("record__amount")))
                  .filter(~Q(balance=0))]
    for record in records:
        record.is_balanced = record.transaction.pk not in imbalanced


def find_order_holes(records: Iterable[Record]) -> None:
    """"Finds whether the order of the transactions on this day is not
        1, 2, 3, 4, 5..., and should be reordered, and sets their
        has_order_holes attributes.

    Args:
        records: The accounting records.
    """
    holes = [x["date"] for x in Transaction.objects
             .values("date")
             .annotate(count=Count("ord"),
                       max=Max("ord"),
                       min=Min("ord"))
             .filter(~(Q(max=F("count")) & Q(min=1)))] +\
            [x["date"] for x in Transaction.objects
             .values("date", "ord")
             .annotate(count=Count("pk"))
             .filter(~Q(count=1))]
    for record in records:
        record.has_order_hole = record.pk is not None\
                                and record.transaction.date in holes


def find_payable_records(account: Account, records: Iterable[Record]) -> None:
    """Finds and sets the whether the payable record is paid.

    Args:
        account: The current ledger account.
        records: The accounting records.
    """
    try:
        payable_accounts = settings.ACCOUNTING["PAYABLE_ACCOUNTS"]
    except AttributeError:
        return
    except TypeError:
        return
    except KeyError:
        return
    if not isinstance(payable_accounts, list):
        return
    if account.code not in payable_accounts:
        return
    rows = Record.objects\
        .filter(
            account__code__in=payable_accounts,
            summary__isnull=False)\
        .values("account__code", "summary")\
        .annotate(
            balance=Sum(Case(When(is_credit=True, then=1), default=-1)
                        * F("amount")))\
        .filter(~Q(balance=0))
    keys = ["%s-%s" % (x["account__code"], x["summary"]) for x in rows]
    for x in [x for x in records
              if x.pk is not None
              and F"{x.account.code}-{x.summary}" in keys]:
        x.is_payable = True


def find_existing_equipments(account: Account,
                             records: Iterable[Record]) -> None:
    """Finds and sets the equipments that still exist.

    Args:
        account: The current ledger account.
        records: The accounting records.
    """
    try:
        equipment_accounts = settings.ACCOUNTING["EQUIPMENT_ACCOUNTS"]
    except AttributeError:
        return
    except TypeError:
        return
    except KeyError:
        return
    if not isinstance(equipment_accounts, list):
        return
    if account.code not in equipment_accounts:
        return
    rows = Record.objects\
        .filter(
            account__code__in=equipment_accounts,
            summary__isnull=False)\
        .values("account__code", "summary")\
        .annotate(
            balance=Sum(Case(When(is_credit=True, then=1), default=-1)
                        * F("amount")))\
        .filter(~Q(balance=0))
    keys = ["%s-%s" % (x["account__code"], x["summary"]) for x in rows]
    for x in [x for x in records
              if x.pk is not None
              and F"{x.account.code}-{x.summary}" in keys]:
        x.is_existing_equipment = True
